/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
 
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ao/ao.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#ifdef HAVE_VORBIS
# include "ogg.h"
#endif
#ifdef HAVE_MAD
# include "mp3.h"
#endif
#include "wav.h"
#include "audio.h"
#include "interface.h"
#include "playlist.h"
#include "filelist.h"
#include "config.h"
#include "options.h"

static short is_playing = 0;
pthread_cond_t paused_cond;
short paused;
short stop_req = 0;
short next_req = 0;
static ao_device *device;
static pthread_t play_tid;
static void (*seek_func)(int ms);
static plist *now_playing = NULL;
static short play_files_running = 0;
static short audio_opened = 0;
#ifdef HAVE_OSS
static int mixer_fd;
#endif

/* Mutexes */
pthread_mutex_t is_playing_mut = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t paused_mut = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t req_mut = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t now_playing_mut = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t play_files_running_mut = PTHREAD_MUTEX_INITIALIZER;

void init_audio () {
	ao_initialize ();
	pthread_cond_init (&paused_cond, NULL);

#ifdef HAVE_OSS

	/* Open mixer device */
	mixer_fd = open ("/dev/mixer", O_RDWR);
	if (mixer_fd == -1) {
		open ("/dev/mixer0", O_RDWR);
	}
	if (mixer_fd == -1) {
		show_err ("Can't open mixer device");
	}
#endif

}

void shutdown_audio () {
	ao_shutdown ();
	
#ifdef HAVE_OSS
	close (mixer_fd);
#endif

}

/* Return 1 on fail */
int open_audio (int bits, int channels, int rate) {
	ao_sample_format format;
	int default_driver;
	
	default_driver = ao_default_driver_id ();
	format.bits = bits;
	format.channels = channels;
	format.rate = rate;
	format.byte_format = AO_FMT_LITTLE;

	device = ao_open_live (default_driver, &format, NULL);
	if (device == NULL) {
		show_err ("Error opening audio device.");
		return 1;
	}

	audio_opened = 1;

	return 0;
}

void close_audio () {
	if (audio_opened)
		ao_close (device);
	audio_opened = 0;
}

void play_buff (char *buff, const int size) {
	ao_play (device, buff, size);
}

void pause_stream () {
	short playing;

	pthread_mutex_lock (&is_playing_mut);
	playing = is_playing;
	pthread_mutex_unlock (&is_playing_mut);

	if (!playing)
		return;
	
	pthread_mutex_lock (&paused_mut);
	paused = 1;
	pthread_mutex_unlock (&paused_mut);
	set_state ("||");
}

void unpause_stream () {
	short playing;

	pthread_mutex_lock (&is_playing_mut);
	playing = is_playing;
	pthread_mutex_unlock (&is_playing_mut);

	if (!playing)
		return;
	
	pthread_mutex_lock (&paused_mut);
	paused = 0;
	pthread_cond_signal (&paused_cond);
	pthread_mutex_unlock (&paused_mut);
	set_state (" >");
}

/* Stop playing the stream */
void stop () {
	short playing;

	unpause_stream ();

	pthread_mutex_lock (&play_files_running_mut);
	playing = play_files_running;
	pthread_mutex_unlock (&play_files_running_mut);
	if (playing) {
		pthread_mutex_lock (&req_mut);
		stop_req = 1;
		pthread_mutex_unlock (&req_mut);
		pthread_join (play_tid, NULL);
		play_files_running = 0;
		stop_req = 0;
	}
	now_playing = NULL;
}

/* Stop playing and go to the next song */
void next () {
	short playing;

	pthread_mutex_lock (&play_files_running_mut);
	playing = play_files_running;
	pthread_mutex_unlock (&play_files_running_mut);
	if (playing) {
		pthread_mutex_lock (&req_mut);
		stop_req = 1;
		next_req = 1;
		pthread_mutex_unlock (&req_mut);
	}
}

void seek_stream (int ms) {
	pthread_mutex_lock (&is_playing_mut);
	if (is_playing)
		seek_func (ms);
	pthread_mutex_unlock (&is_playing_mut);
}

#ifdef HAVE_VORBIS
/* Start playing an .ogg file in another thread */
void play_ogg (char *file_name) {
	seek_func = seek_ogg;
	if (!load_ogg (file_name)) {
		pthread_mutex_lock (&is_playing_mut);
		is_playing = 1;
		pthread_mutex_unlock (&is_playing_mut);

		process_ogg ();

		pthread_mutex_lock (&is_playing_mut);
		is_playing = 0;
		pthread_mutex_unlock (&is_playing_mut);
	}
}
#endif

#ifdef HAVE_MAD
/* Start playing an .mp3 file in another thread */
void play_mp3 (char *file_name) {
	seek_func = seek_mp3;
	if (!load_mp3 (file_name)) {
		pthread_mutex_lock (&is_playing_mut);
		is_playing = 1;
		pthread_mutex_unlock (&is_playing_mut);

		process_mp3 ();

		pthread_mutex_lock (&is_playing_mut);
		is_playing = 0;
		pthread_mutex_unlock (&is_playing_mut);
	}
}
#endif

/* Start playing a .wav file in another thread */
void play_wav (char *file_name) {
	seek_func = seek_wav;
	if (!load_wav (file_name)) {
		pthread_mutex_lock (&is_playing_mut);
		is_playing = 1;
		pthread_mutex_unlock (&is_playing_mut);

		process_wav ();

		pthread_mutex_lock (&is_playing_mut);
		is_playing = 0;
		pthread_mutex_unlock (&is_playing_mut);
	}
}

/* Play files from the playlist */
void *play_files (void *unused) {
	short stop;
	short next;
	plist *playing;
	short shuffle;
	
	pthread_mutex_lock (&now_playing_mut);
	playing = now_playing;
	pthread_mutex_unlock (&now_playing_mut);
	
	while (playing) {
		reset_file_info ();
		set_state (" >");
		mark_item (playing->item.file_name);

		/* Set the title */
		LOCK_HLIST ();
		if (playing->item.info)
			set_title (playing->item.title);
		else {
			char *title;
			song_info *info;

			info = get_song_info (playing->item.file_name,
					read_file_type (playing->item.file_name));
			if (info) {
				title = build_title (info);
				if (title) {
					set_title (title);
					free (title);
				}
				else
					set_title (short_name(playing->item.file_name));
				free_song_info (info);
			}
			else
				set_title (playing->item.title);
		}
		UNLOCK_HLIST ();
				
#ifdef HAVE_VORBIS
		if (playing->item.type == F_OGG)
			play_ogg (playing->item.file_name);
#endif
#ifdef HAVE_MAD		
		if (now_playing->item.type == F_MP3)
			play_mp3 (playing->item.file_name);
#endif
		if (now_playing->item.type == F_WAV)
			play_wav (playing->item.file_name);
		
		set_title ("");
		unmark_item ();
		set_state ("[]");
		set_curr_time (0);

		/* If playing was stopped by the user, stop playing the
		 * playlist
		 */
		pthread_mutex_lock (&req_mut);
		stop = stop_req;
		next = next_req;
		stop_req = 0;
		next_req = 0;
		pthread_mutex_unlock (&req_mut);
		if (stop && !next)
			break;

		OPTIONS_LOCK ();
		shuffle = options.shuffle;
		OPTIONS_UNLOCK ();

		pthread_mutex_lock (&now_playing_mut);
		if (shuffle) {
			shuffled_list *sh_item;
			
			sh_item = find_shuffled (shuffled, now_playing);
			if (sh_item && sh_item->next)
				now_playing = sh_item->next->item;
			else
				now_playing = NULL;
		}
		else {
			if (now_playing)
				now_playing = now_playing->next;
		}

		OPTIONS_LOCK ();
		if (options.repeat && !now_playing) {
			if (options.shuffle)
				now_playing = shuffled->item;
			else
				now_playing = hidden_list->next;
		}
		OPTIONS_UNLOCK ();
		
		playing = now_playing;
		pthread_mutex_unlock (&now_playing_mut);
	}
	
	return NULL;
}

/* Play file. Shuffled is a shuffled list or NULL if there is no such list. */
void play_it (plist *item) {
	now_playing = item;
	play_files_running = 1;
	if (pthread_create (&play_tid, NULL, play_files, NULL))
			fatal ("pthread_create() failed!");
}

/* Get the file name of the currently played file */
plist *get_played_item () {
	plist *item;
	
 	pthread_mutex_lock (&now_playing_mut);
	if (now_playing != NULL)
		item = now_playing;
	else
		item = NULL;
	pthread_mutex_unlock (&now_playing_mut);

	return item;
}

/* Return value != 0 if playing something */
short playing () {
	short playing;

	pthread_mutex_lock (&is_playing_mut);
	playing = is_playing;
	pthread_mutex_unlock (&is_playing_mut);

	return playing;
}

#ifdef HAVE_OSS

/* Get PCM volume, return -1 on error */
int read_mixer () {
	int vol;
	
	if (mixer_fd != -1) {
		if (ioctl(mixer_fd, MIXER_READ(SOUND_MIXER_PCM), &vol) == -1)
			show_err ("Can't read from mixer");
		else {
			/* Average between left and right */
			return ((vol & 0xFF) + ((vol >> 8) & 0xFF)) / 2;
		}
	}

	return -1;
}

/* Set PCM volume */
void set_mixer (int vol) {
	if (mixer_fd != -1) {
		if (vol > 100)
			vol = 100;
		else if (vol < 0)
			vol = 0;
		
		vol = vol | (vol << 8);
		if (ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_PCM), &vol) == -1)
			show_err ("Can't set mixer");
	}
}

#endif
