/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
 
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

/* Include dirent for various systems */
#ifdef HAVE_DIRENT_H
# include <dirent.h>
#else
# define dirent direct
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
#endif

#include "audio.h"
#include "main.h"
#include "interface.h"
#include "playlist.h"
#include "config.h"
#ifdef HAVE_MAD
# include "mp3.h"
#endif
#ifdef HAVE_VORBIS
#include "ogg.h"
#endif
#include "options.h"

plist_menu *file_menu = NULL;
int counter = 0; /* Counter for progress bar */

/* Check type of the file */
short read_file_type (const char *name) {
	struct stat file_stat;

	if (stat(name, &file_stat) == -1)
		return F_OTHER; /* Ignore the file if stat() failed */

	/* Directory */
	if (S_ISDIR(file_stat.st_mode))
		return F_DIRECTORY;

	/* The name must by at least 4 chars long to have an extension */
	if (strlen(name) < 4)
		return F_OTHER;

#ifdef HAVE_VORBIS
	/* OGG */
	if (!(strcasecmp(name + strlen(name) - 4, ".OGG")))
		return F_OGG;
#endif

#ifdef HAVE_MAD
	/* MP3 */
	if (!(strcasecmp(name + strlen(name) - 4, ".MP3")))
		return F_MP3;
#endif
	
	/* WAV */
	if (!(strcasecmp(name + strlen(name) - 4, ".WAV")))
		return F_WAV;

	/* Not recognized */
	return F_OTHER;
}

/* Get the name of the current directory */
char *get_dir () {
	static char abs_dir[PATH_MAX];
	char *dir;

	getcwd (abs_dir, sizeof(abs_dir));
	dir = strrchr (abs_dir, '/');
	if (dir == NULL)
		return abs_dir;
	else
		return dir + 1;
}

/* Make an absolute path from getcwd() and file name, return NULL on error
 * or pointer to buff. buff must be at least PATH_MAX big
 */
char *abs_path (const char *file, char *buff) {
	char cwd[PATH_MAX];
	char *c;
	int file_pos = 0;
	int file_len = strlen(file);

	/* The file is already an absolute path */
	if (file[0] == '/') {
		strcpy (buff, file);
		return buff;
	}

	if (getcwd(cwd, sizeof(cwd)) == NULL)
		return NULL;

	while (1) {

		/* Skip './' */
		if (file_len - file_pos > 2
				&& file[file_pos] == '.'
				&& file[file_pos+1] == '/')
			file_pos += 2;

		/* Remove one directory if '../' */
		else if (file_len - file_pos > 3
				&& file[file_pos] == '.'
				&& file[file_pos+1] == '.'
				&& file[file_pos+2] == '/') {
			file_pos += 3;
			c = strrchr (cwd, '/');
			if (c == NULL)
				return NULL;
			*c = 0;
		}
		else break;
	}

	strcpy (buff, cwd);
	strcat (buff, "/");
	strcat (buff, file + file_pos);
	
	return buff;
}

song_info *get_song_info (const char *file_name, const short type) {
	song_info *info;

	if (type == F_OTHER || type == F_DIRECTORY || type == F_WAV)
		return NULL;

	info = (song_info *)cmalloc (sizeof(song_info));
	info->title = NULL;
	info->artist = NULL;
	info->album = NULL;
	
	switch (type) {
#ifdef HAVE_MAD
		case F_MP3:
			read_mp3_info (file_name, info);
			break;
#endif
#ifdef HAVE_VORBIS
		case F_OGG:
			read_ogg_info (file_name, info);
			break;
#endif
	}
	
	return info;
}

/* Build a title based on song_info, you must free retrivied value. */
char *build_title (const song_info *info) {
	char tmp_title[100] = "";
	char *title;
	int len = sizeof (tmp_title) - 1;

	/* We MUST use if (len > 0) because argument to strn*() is size_t,
	 * which can be unsigned, so our -3 will be very big number after the
	 * cast.
	 */

	if (info->artist != NULL) {
		strncpy (tmp_title, info->artist, len);
		len -= strlen (info->artist);
	}
	if (info->title != NULL && len > 0) {
		strncat (tmp_title, " - ", len);
		len -= 3;
		
		if (len > 0) {
			strncat (tmp_title, info->title, len);
			len -= strlen (info->title);
		}
	}
	if (info->album != NULL && len > 0) {
		strncat (tmp_title, " (", len);
		len -= 2;

		if (len > 0) {
			strncat (tmp_title, info->album, len);
			len -= strlen (info->album);

			if (len > 0)
				strncat (tmp_title, ")", len);
		}
	}

	if (tmp_title[0] == 0)
		return NULL;
	
	tmp_title[sizeof(tmp_title)-1] = 0;

	title = strdup (tmp_title);
	
	return title;
}

/* Return short file name */
char *short_name (char *name) {
	char *sh;

	sh = strrchr (name, '/');
	if (sh)
		return sh + 1;
	else
		return name;
}

/* Add file to the play list useing add_to_plist() */
void add_file (plist *list, char *file_name, const short sort) {
	int len;
	short type;
	char abs_file_name[PATH_MAX];
	char my_name[PATH_MAX+1];
	song_info *info;
	char *title = NULL;

	/* Update progress bar */
	if (counter++ % 10 == 0)
		move_progress_bar ();


	type = read_file_type (file_name);
	if (type == F_OTHER)
		return;

	/* Add '/' at the end of the directory name */
	if (type == F_DIRECTORY) {
		strncpy (my_name, short_name(file_name), sizeof(my_name) - 1);
		len = strlen(my_name);
		my_name[len] = '/';
		my_name[len + 1] = 0;
	}
	
	else {
		strncpy (my_name, file_name, sizeof(my_name));
	}
	
	if (abs_path(file_name, abs_file_name) == NULL)
		strcpy (abs_file_name, my_name);

	if (type != F_DIRECTORY) {
		if (options.full_info) {
			info = get_song_info (file_name, type);
			if (info)
				title = build_title (info);
		}
		else {
			title = NULL;
			info = NULL;
		}
	
		add_to_plist (list, abs_file_name,
				title ? title : short_name(my_name),
				info, type, sort);
		if (title)
			free (title);
	}
	else
		add_to_plist (list, abs_file_name, my_name, NULL,
				type, sort);

}

/* Add directory to a list, return 0 on success */
int add_dir (const char *name, plist *list, short recurrence) {
	DIR *dir;
	struct dirent *entry;

	if (!(dir = opendir (name))) {
		show_err ("opendir() failed: %s", strerror(errno));
		return 1;
	}
	
	while ((entry = readdir (dir)) != NULL) {
		char path[PATH_MAX];

		snprintf (path, sizeof(path), "%s/%s", name, entry->d_name);

		if (!strcmp (entry->d_name, "."))
			continue;
		if (recurrence) {
			if (read_file_type (path) == F_DIRECTORY) {
				if (strcmp(entry->d_name, ".."))
					add_dir (path, list, 1);
			}
			else
				add_file (list, path, 0);
		}
		else
			add_file (list, path, 1);
	}

	closedir (dir);
	return 0;
}

int enter_dir (const char *dir_name) {
	char abs_dir_name[PATH_MAX];

	/* Go to the directory */
	if (chdir(dir_name) == -1) {
		show_err ("chdir() failed: %s", strerror(errno));
		return 1;
	}

	tmp_list = new_plist ();
	
	if (add_dir (".", tmp_list, 0)) {
		chdir ("..");
		free_plist (tmp_list);
		return 1;
	}

	getcwd (abs_dir_name, sizeof(abs_dir_name));	
	main_title (abs_dir_name);

	if (file_menu)
		free_plist_menu (file_menu);

	curr_plist = tmp_list;
	
	/* Make the menu */
	file_menu = make_plist_menu (tmp_list);
	set_num_items (file_menu->menu->nitems);

	return 0;
}

/* Make titles for menu items */
void rebuild_menu_titles (plist *playlist) {
	plist *list_item;
	char *title;

	for (list_item = playlist->next; list_item != NULL;
			list_item = list_item->next) {
		if (list_item->item.type != F_DIRECTORY) {
			if (list_item->item.title)
				free (list_item->item.title);
		
			if (options.full_info) {
				if (!list_item->item.info)
					list_item->item.info = get_song_info (
							list_item->item.file_name,
							read_file_type (list_item->item.file_name));
				
				title = NULL;
				if (list_item->item.info)
					title = build_title (list_item->item.info);
				if (!title)
					title = strdup (short_name
							(list_item->item.file_name));

				list_item->item.title = title;
				list_item->item.menu_item->title = title;
			}
			else {
				title = short_name (list_item->item.file_name);
				list_item->item.title = strdup (title);
				list_item->item.menu_item->title =
					list_item->item.title;
			}
		}
	}
}
