/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <sys/types.h>
#ifdef VERSION_CHECKER
# include <sys/wait.h>
#endif
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ncurses.h>
#include <pthread.h>
#include "audio.h"
#include "filelist.h"
#include "interface.h"
#include "config.h"
#include "main.h"
#include "options.h"
#ifdef VERSION_CHECKER
# include "version_checker.h"
#endif

/* Mode in which is the main window */
enum {
	WIN_MENU,
	
#ifdef VERSION_CHECKER
	WIN_VER_CHECKER,
#endif

	WIN_NOCURSES,
	WIN_HELP
};
static int main_win_mode = WIN_MENU;

WINDOW *main_win = NULL;
WINDOW *main_subwin;

static WINDOW *info_win;
short colors;

static struct file_info_s {
	char title[100];
	char bitrate[4];
	char rate[3];
	char time[6];
	char curr_time[6];
	short channels;
	char state[3];
} file_info;

static time_t err_timeout = 0; /* Timeout for showing error message */
static char err_msg[100];
static char mainwin_title[100] = PACKAGE_NAME;
static char old_title[100];
static char status_line[11] = "          ";
static int marked = -1; /* Marked item */
static short buggy_xterm = 0;
plist *curr_plist;

/* Progress bar */
struct progress_bar progress_bar = {
	{ 2, 1, 0 },
	{ 1, 1, 1},
	0
};

/* Lock the screen to disallow more than one thread printing to the screen.
 * Also, it is used to lock the file_info structure.
 */
pthread_mutex_t scr_mut = PTHREAD_MUTEX_INITIALIZER;

/* End program with message */
void fatal (const char *msg) {
	shutdown_audio ();
	endwin ();
	fprintf (stderr, "%s\n", msg);
	exit (1);
}

/* Draw border around the main window */
static void main_border () {
	LOCK_SCR ();
	wborder (main_win, ACS_VLINE, ACS_VLINE, ACS_HLINE, ' ',
			ACS_ULCORNER, ACS_URCORNER, ACS_VLINE, ACS_VLINE);

	/* The title */
	wmove (main_win, 0, COLS / 2 - strlen(mainwin_title) / 2 - 1);
	waddch (main_win, ACS_RTEE);
	waddstr (main_win, mainwin_title);
	waddch (main_win, ACS_LTEE);

	UNLOCK_SCR ();
}

/* Draw border around the info window and some text */
static void info_border () {
	LOCK_SCR ();
	werase (info_win);
	wborder (info_win, ACS_VLINE, ACS_VLINE, ACS_HLINE, ACS_HLINE,
			ACS_LTEE, ACS_RTEE, ACS_LLCORNER, ACS_LRCORNER);
	wmove (info_win, 2, 25);
	waddstr (info_win, "KHz    Kbps");
	UNLOCK_SCR ();
}

/* Reset the title, bitrate etc */
void reset_file_info () {
	LOCK_SCR ();
	file_info.title[0] = 0;
	file_info.time[0] = 0;
	file_info.curr_time[0] = 0;
	file_info.bitrate[0] = 0;
	file_info.rate[0] = 0;
	file_info.channels = 1;
	UNLOCK_SCR ();
	set_state ("[]");
}

static void refresh_info_win () {
	if (main_win_mode != WIN_NOCURSES)
		wrefresh (info_win);
}

#ifdef HAVE_OSS

/* Draw mixer bar */
static void draw_mixer () {
	char bar[21];
	int vol;

	vol = read_mixer ();
	if (vol == -1)
		return;

	if (vol == 100)
		sprintf (bar, " Vol %d%%           ", vol);
	else
		sprintf (bar, " Vol  %02d%%           ", vol);
	mvwaddch (info_win, 0, COLS - 23, ACS_RTEE);
	mvwaddch (info_win, 0, COLS - 2, ACS_LTEE);

	wattron (info_win, COLOR_PAIR(6));
	mvwaddnstr (info_win, 0, COLS - 22, bar, vol / 5);
	wattron (info_win, COLOR_PAIR(1));
	mvwaddnstr (info_win, 0, COLS - 22 + (vol / 5),
			bar + vol / 5, 20 - (vol / 5));
}

#endif

static void draw_progress_bar () {
	int i;

	mvwaddstr (info_win, 0, 6, "          ");
	for (i = 2; i >= 0; i--) {
		wattron (info_win, COLOR_PAIR(7 + i) | A_BOLD);
		mvwaddch (info_win, 0, 6 + progress_bar.pos[i], '-');
	}
}

/* Update the info win */
void update_info_win () {
	info_border ();
	LOCK_SCR ();
	wattron (info_win, A_BOLD);

	/* Show error message it it didn't expire yet */
	if (time(NULL) <= err_timeout) {
		if (colors)
			wattron (info_win, COLOR_PAIR(3));
		mvwaddstr (info_win, 1, 1, err_msg);
		if (colors)
			wattroff (info_win, COLOR_PAIR(3));
	}
	else {

		/* The title */
		mvwaddstr (info_win, 1, 4, file_info.title);

		/* State of playing */
		mvwaddstr (info_win, 1, 1, file_info.state);
	}

	/* Current time */
	wmove (info_win, 2, 6);
	waddstr (info_win, file_info.curr_time);

	/* Total_time */
	if (file_info.time[0] != 0) {
		wmove (info_win, 2, 12);
		wprintw (info_win, "[%s]", file_info.time);
	}

	/* Rate */
	wmove (info_win, 2, 25 - strlen(file_info.rate));
	waddstr (info_win, file_info.rate);
	
	/* Bitrate */
	wmove (info_win, 2, 32 - strlen(file_info.bitrate));
	waddstr (info_win, file_info.bitrate);

	/* Channels */
	wmove (info_win, 2, 38);
	if (file_info.channels == 2)
		wattron (info_win, COLOR_PAIR(1) | A_BOLD);
	else
		wattron (info_win, COLOR_PAIR(9) | A_BOLD);
	waddstr (info_win, "[STEREO]");
	
	/* Shuffle & repeat */
	wmove (info_win, 2, COLS - sizeof("[SHUFFLE] [REPEAT] "));
	OPTIONS_LOCK ();
	if (options.shuffle)
		wattron (info_win, COLOR_PAIR(1) | A_BOLD);
	else
		wattron (info_win, COLOR_PAIR(9) | A_BOLD);
	waddstr (info_win, "[SHUFFLE] ");

	if (options.repeat)
		wattron (info_win, COLOR_PAIR(1) | A_BOLD);
	else
		wattron (info_win, COLOR_PAIR(9) | A_BOLD);
	
	OPTIONS_UNLOCK ();
	waddstr (info_win, "[REPEAT]");

	wattron (info_win, COLOR_PAIR(1));


	/* Status line */
	wattroff (info_win, A_BOLD);
	mvwaddch (info_win, 0, 5, ACS_RTEE);
	if (progress_bar.on)
		draw_progress_bar ();
	else
		mvwaddstr (info_win, 0, 6, status_line);
	wattroff (info_win, A_BOLD);
	mvwaddch (info_win, 0, 16, ACS_LTEE);
	
#ifdef HAVE_OSS
	draw_mixer ();
#endif

	refresh_info_win ();

	UNLOCK_SCR ();
}

void set_num_items (int num) {
	LOCK_SCR ();
	sprintf (status_line, "%10d", num);
	UNLOCK_SCR ();
	update_info_win ();
}

/* Init ncurses, make windows */
void init_interface () {
	initscr ();
	colors = has_colors () == TRUE ? 1 : 0;
	cbreak ();
	noecho ();

	if (COLS < 70 || LINES < 20)
		fatal ("Your terminal is too small for this program.");
	
	if (colors) {
		start_color ();
		init_pair (1, COLOR_WHITE, COLOR_BLUE); /* Menu item */
		init_pair (2, COLOR_WHITE, COLOR_BLACK); /* Selected menu item */
		init_pair (3, COLOR_RED, COLOR_BLUE); /* Errors */
		init_pair (4, COLOR_GREEN, COLOR_BLUE); /* Marked item */
		init_pair (5, COLOR_GREEN, COLOR_BLACK); /* Marked and selected */
		init_pair (6, COLOR_BLACK, COLOR_CYAN); /* Volume bar */

		/* Colors for the status line */
		init_pair (7, COLOR_WHITE, COLOR_BLUE);
		init_pair (8, COLOR_GREEN, COLOR_BLUE);
		init_pair (9, COLOR_BLUE, COLOR_BLUE);
	}
	refresh ();

	reset_file_info ();

	/* Try to detect if we are running in buggy X terminal */
	if (getenv("DISPLAY") && strcmp(getenv("TERM"), "xterm"))
		buggy_xterm = 1;
	
	/* Main window */
	main_win = newwin (LINES - 4, COLS, 0, 0);
	keypad (main_win, TRUE);
	if (colors)
		wbkgd (main_win, COLOR_PAIR(1));
	main_border ();
	wrefresh (main_win);

	/* Info window */
	info_win = newwin (4, COLS, LINES - 4, 0);
	if (colors)
		wbkgd (info_win, COLOR_PAIR(1));
	update_info_win ();
}

/* Free all interface structures and end ncurses */
void end_interface () {
	free_plist_menu (file_menu);
	free_plist (tmp_list);
	free_plist (playlist);
	if (hidden_list != tmp_list && hidden_list != playlist)
		free_plist (hidden_list);
	endwin ();
}

/* Show errr message in the info window */
void show_err (const char *format, ...) {
	va_list va;
	
	err_timeout = time(NULL) + 3;
	va_start (va, format);
	LOCK_SCR ();
	vsnprintf (err_msg, 99, format, va);
	err_msg[99] = 0;
	UNLOCK_SCR ();

	if (main_win)
		update_info_win ();

	va_end (va);
}

/* Change the title of the main window */
void main_title (const char *title) {
	strncpy (mainwin_title, title, sizeof(mainwin_title));
	mainwin_title[sizeof(mainwin_title)-1] = 0;
	main_border ();
	LOCK_SCR ();
	wrefresh (main_win);
	UNLOCK_SCR ();
}

/* Set title of the song to show */
void set_title (const char *title) {
	LOCK_SCR ();
	strncpy (file_info.title, title, COLS - 5);
	file_info.title[99] = 0;
	UNLOCK_SCR ();
	update_info_win ();
}

/* Set bitrate to show */
void set_bitrate (const int bitrate) {
	LOCK_SCR ();
	if (bitrate == 0)
		strcpy (file_info.bitrate, "--");
	else {
		snprintf (file_info.bitrate, 4, "%d", bitrate);
		file_info.bitrate[3] = 0;
	}
	UNLOCK_SCR ();
	update_info_win ();
}

/* Set rate to show */
void set_rate (const int rate) {
	LOCK_SCR ();
	if (rate == 0)
		strcpy (file_info.rate, "--");
	else {
		snprintf (file_info.rate, 3, "%d", rate);
		file_info.rate[2] = 0;
	}

	UNLOCK_SCR ();
	update_info_win ();
}

/* Set state (play, stop or pause) */
void set_state (const char *state) {
	LOCK_SCR ();
	strcpy (file_info.state, state);
	UNLOCK_SCR ();
	update_info_win ();
}

/* Convert time in second to min:sec text format */
static void sec_to_min (char *buff, const long seconds) {
	int min, sec;

	min = seconds / 60;
	sec = seconds % 60;

	snprintf (buff, 6, "%02d:%02d", min, sec);
}

/* Set the current time of sream */
void set_curr_time (const double time) {
	LOCK_SCR ();
	sec_to_min (file_info.curr_time, time);
	UNLOCK_SCR ();
	update_info_win ();
}

/* Set the total time of the stream */
void set_time (const double time) {
	LOCK_SCR ();
	sec_to_min (file_info.time, time);
	UNLOCK_SCR ();
	update_info_win ();
}

/* Show info whether srteam is stereo or mono */
void set_stereo (const short stereo) {
	LOCK_SCR ();
	file_info.channels = stereo == 1 ? 2 : 1;
	UNLOCK_SCR ();
	update_info_win ();
}

/* Mark item with color */
void mark_item (const char *file) {
	int i;
	plist *plist_item;

	LOCK_SCR ();
	for (i = 0; file_menu->items[i] != NULL; i++) {
		plist_item = (plist *)file_menu->items[i]->usrptr;
		if (!strcmp(plist_item->item.file_name, file)) {
			file_menu->items[i]->attr_normal = COLOR_PAIR(4) | A_BOLD;
			file_menu->items[i]->attr_sel = COLOR_PAIR(5) | A_BOLD;
			marked = i;
			break;
		}
	}
	menu_draw (file_menu->menu);
	wrefresh (main_win);
	UNLOCK_SCR ();
}

/* Unmark item */
void unmark_item () {
	if (marked != -1) {
		LOCK_SCR ();
		if (((plist*)(file_menu->items[marked]->usrptr))->item.type
				!= F_DIRECTORY) {
			file_menu->items[marked]->attr_normal = COLOR_PAIR(1);
			file_menu->items[marked]->attr_sel = COLOR_PAIR(2);
		}
		else {
			file_menu->items[marked]->attr_normal = COLOR_PAIR(1)
				| A_BOLD;
			file_menu->items[marked]->attr_sel = COLOR_PAIR(2)
				| A_BOLD;
		}
		menu_draw (file_menu->menu);
		wrefresh (main_win);
		UNLOCK_SCR ();
	}
}

static void mark_played () {
	plist *played;

	marked = -1;
	played = get_played_item ();
	
	if (played)
		mark_item (played->item.file_name);
}

/* Go to the specified directory */
void go_to_dir (const char *name) {
	int num;
	short show_last = 0;
	int len = strlen (name);
	static char last_dir[PATH_MAX+1];
	plist *old_list;

	LOCK_SCR ();
	progress_bar.on = 1;
	UNLOCK_SCR ();
	update_info_win ();

	if (name[len-2] == '.' && name[len-1] == '.')
		show_last = 1;
	
	/* Get current directory and add '/' as in the item titles */
	strncpy (last_dir, get_dir(), sizeof(last_dir) - 1);
	len = strlen (last_dir);
	last_dir[len] = '/';
	last_dir[len+1] = 0;
	
	old_list = tmp_list;
	if (!enter_dir(name)) {
		if (old_list != hidden_list)
			free_plist (old_list);
	}
	else {
		tmp_list = old_list;
	}
		
	/* Make previous directory selected when entering to '../' */
	if (show_last) {
		num = menu_find_item (file_menu->menu, last_dir);
		if (num != -1)
			menu_setcurritem (file_menu->menu, num);
	}

	mark_played ();

	LOCK_SCR ();
	progress_bar.on = 0;
	UNLOCK_SCR ();
	update_info_win ();
}

/* Take an action on a file selected by the menu */
void to_file () {
	plist *item;
	MenuItem *mitem;
		
	mitem = menu_curritem (file_menu->menu);
	item = (plist *)mitem->usrptr;

	/* Stop showing error message */
	err_timeout = 0;
	update_info_win ();
	
	if (item->item.type == F_DIRECTORY)
		go_to_dir (item->item.file_name);
	else if (item->item.type != F_OTHER) {
		stop ();
		
		if (hidden_list != curr_plist && hidden_list != playlist
				&& hidden_list)
			free_plist (hidden_list);
		hidden_list = curr_plist;

		if (options.shuffle) {
			if (shuffled)
				free_shuffled_list (shuffled);
			shuffled = gen_shuffled_list (curr_plist, item);
			play_it (item);
		}
		else
			play_it (item);
	}
}

/* Show help window */
static void show_help () {
	strcpy (old_title, mainwin_title);
	main_title ("HELP");

	LOCK_SCR ();
	main_win_mode = WIN_HELP;

	main_subwin = subwin (main_win, LINES - 6, COLS - 2, 1, 1);
	if (colors)
		wbkgd (main_subwin, COLOR_PAIR(1));
	werase (main_subwin);

	mvwprintw (main_subwin, 0, 0, "\
  UP, DOWN	Move up and down in the menu
  PAGE UP/DOWN	Move one page up/down
  HOME, END	Move to the first, last item
  ENTER		Start playing files (from this file) or go to directory
  s		Stop playing
  n		Next song
  p		Pause/unpause
  LEFT, RIGHT	Seek backward, forward
  h		Show this help screen
  f		Switch between short and full names
  m		Go to the music directory (requires an entry in the config)
  r		Redraw the screen
  a/A		Add file to the playlist / Add directory recursively
  d/C		Delete item from the playlist / Clear the playlist
  l		Switch between playlist and file list
  S/R		Switch shuffle / repeat
");

#ifdef VERSION_CHECKER
	wprintw (main_subwin, "\
  v		Check if a newer version of the MOC is available
");
#endif
	
#ifdef HAVE_OSS
	wprintw (main_subwin, "\
  '.' , ','	Increase, decrease volume by 5%
  '>' , '<'	Increase, decrease volume by 1%
");
#endif

	wrefresh (main_subwin);
	wrefresh (main_win);

	UNLOCK_SCR ();
}

#ifdef VERSION_CHECKER

/* Switch main window into text mode */
void win_text_mode () {
	strcpy (old_title, mainwin_title);
	main_title ("Version checker");

	LOCK_SCR ();
	main_win_mode = WIN_VER_CHECKER;

	main_subwin = subwin (main_win, LINES - 6, COLS - 2, 1, 1);
	if (colors)
		wbkgd (main_subwin, COLOR_PAIR(1));
	werase (main_subwin);
	wrefresh (main_subwin);
	wrefresh (main_win);

	UNLOCK_SCR ();
}

#endif

/* Redraw the screen */
static void refresh_screen () {
	LOCK_SCR ();
	wclear (main_win);
	wclear (info_win);
	UNLOCK_SCR ();
	main_border ();
	LOCK_SCR ();
	menu_draw (file_menu->menu);
	wrefresh (main_win);
	UNLOCK_SCR ();
	update_info_win ();
}

/* Switch to playlist or file list */
static void switch_list () {
	if (curr_plist == tmp_list) {
		if (playlist && playlist->next) {
			curr_plist = playlist;
			main_title ("PLAYLIST");
		}
		else
			show_err ("The playlist is empty.");
	}
	else {
		char abs_dir_name[PATH_MAX];

		/* Sometimes we must make this list first */
		if (!tmp_list)
			enter_dir (curr_dir());
		curr_plist = tmp_list;

		getcwd (abs_dir_name, sizeof(abs_dir_name));	
		main_title (abs_dir_name);
	}

	if (file_menu)
		free_plist_menu (file_menu);
	file_menu = make_plist_menu (curr_plist);

	set_num_items (file_menu->menu->nitems);
	LOCK_SCR ();
	werase (main_win);
	UNLOCK_SCR ();
	main_border ();
	LOCK_SCR ();
	werase (main_win);
	menu_draw (file_menu->menu);
	wrefresh (main_win);
	UNLOCK_SCR ();

	mark_played ();
}

/* Add item to the playlist */
static void append_item () {
	plist *item;
	MenuItem *mitem;

	if (curr_plist == playlist) {
		show_err ("Can't add a file to the playlist from the playlist.");
		return;
	}
		
	mitem = menu_curritem (file_menu->menu);
	item = (plist *)mitem->usrptr;
	
	if (item->item.type != F_DIRECTORY) {
		if (playlist == NULL)
			playlist = new_plist ();
		add_plist_item (playlist, dup_item(item), 0);
	}
	else
		show_err ("To add a directory, use the 'A' command.");
}

/* Add recursively all files in the directory to the playlist */
static void append_dir () {
	plist *item;
	MenuItem *mitem;

	if (curr_plist == playlist) {
		show_err ("Can't add a file to the playlist from the playlist.");
		return;
	}
		
	mitem = menu_curritem (file_menu->menu);
	item = (plist *)mitem->usrptr;

	if (playlist == NULL)
		playlist = new_plist ();

	LOCK_SCR ();
	progress_bar.on = 1;
	UNLOCK_SCR ();
	update_info_win ();

	/* Add selected directory */
	if (item->item.type == F_DIRECTORY &&
			strcmp (item->item.title, "../")) {
		plist *new_list = new_plist ();
		add_dir (item->item.file_name, new_list, 1);
		append_plist (playlist, new_list);
	}

	/* Add current directory */
	else
		append_plist (playlist, dup_plist(curr_plist));

	LOCK_SCR ();
	progress_bar.on = 0;
	UNLOCK_SCR ();
	update_info_win ();
}

/* Delete item from the playlist */
static void delete_item () {
	plist *item;
	MenuItem *mitem;
	int menu_top, menu_sel;

	if (curr_plist != playlist) {
		show_err ("You can delete items only from the playlist.");
		return;
	}

	if (!playlist->next)
		return;

	menu_top = file_menu->menu->top;
	menu_sel = file_menu->menu->selected;

	mitem = menu_curritem (file_menu->menu);
	item = (plist *)mitem->usrptr;

	if (item == get_played_item()) {
		show_err ("Can't delete item that is currently being played.");
		return;
	}

	LOCK_HLIST ();
	delete_plist_item (curr_plist, item);
	UNLOCK_HLIST ();

	if (file_menu)
		free_plist_menu (file_menu);
	file_menu = make_plist_menu (curr_plist);
	set_num_items (file_menu->menu->nitems);

	set_menu_state (file_menu->menu, menu_sel, menu_top);
	mark_played ();
}

static void clear_playlist () {
	if (playlist) {
		if (curr_plist == playlist)
			switch_list ();

		if (hidden_list != playlist)
			free_plist (playlist);
		playlist = NULL;
		show_err ("The playlist was cleared.");
	}
}

/* Handle keys in menu mode, return 1 on quit */
static int menu_command (const int ch) {
	switch (ch) {
		case KEY_DOWN:
			menu_driver (file_menu->menu, REQ_DOWN);
			break;
		case KEY_UP:
			menu_driver (file_menu->menu, REQ_UP);
			break;
		case KEY_NPAGE:
			menu_driver (file_menu->menu, REQ_PGDOWN);
			break;
		case KEY_PPAGE:
			menu_driver (file_menu->menu, REQ_PGUP);
			break;
		case KEY_HOME:
			menu_driver (file_menu->menu, REQ_TOP);
			break;
		case KEY_END:
			menu_driver (file_menu->menu, REQ_BOTTOM);
			break;
		case 's':
			stop ();
			break;
		case 'n':
			next ();
			break;
		case KEY_RIGHT:
			seek_stream (1);
			break;
		case KEY_LEFT:
			seek_stream (-1);
			break;
		case 'p':
			if (!paused)
				pause_stream ();
			else
				unpause_stream ();
			break;
		case '\n':
			to_file ();
			
			/* Some xterms (gnome, kde terms) needs wclear to
			 * correctly refresh the screen.
			 */
			if (buggy_xterm)
				wclear (main_win);
				
			break;
		case 'h':
			show_help ();
			break;
		case 'f':
			options.full_info = !options.full_info;
			LOCK_HLIST ();
			rebuild_menu_titles (curr_plist);
			UNLOCK_HLIST ();
			menu_draw (file_menu->menu);
			break;
		case 'm':
			if (options.music_dir)
				go_to_dir (options.music_dir);
			break;
		case 'r':
			refresh_screen ();
			break;
		case 'l':
			switch_list ();
			break;
		case 'a':
			append_item ();
			break;
		case 'A':
			append_dir ();
			break;
		case 'd':
			delete_item ();
			break;
		case 'C':
			clear_playlist ();
		case 'S':
			OPTIONS_LOCK ();
			options.shuffle = !options.shuffle;
			OPTIONS_UNLOCK ();
			update_info_win ();
			
			if (playing()) {
				if (shuffled)
					free_shuffled_list (shuffled);
				shuffled = gen_shuffled_list (hidden_list,
						get_played_item());
			}
			
			break;
		case 'R':
			OPTIONS_LOCK ();
			options.repeat = !options.repeat;
			OPTIONS_UNLOCK ();
			update_info_win ();
			break;

#ifdef VERSION_CHECKER
		case 'v':
			win_text_mode ();
			check_version ();
			break;
#endif
		case 'q':
			return 1;
#ifdef HAVE_OSS
		case '<':
			set_mixer (read_mixer() - 1);
			update_info_win ();
			break;
		case ',':
			set_mixer (read_mixer() - 5);
			update_info_win ();
			break;
		case '.':
			set_mixer (read_mixer() + 5);
			update_info_win ();
			break;
		case '>':
			set_mixer (read_mixer() + 1);
			update_info_win ();
			break;
#endif
	}

	if (main_win_mode == WIN_MENU) {
		LOCK_SCR ();
		werase (main_win);
		UNLOCK_SCR ();
		main_border ();
		LOCK_SCR ();
		menu_draw (file_menu->menu);
		wrefresh (main_win);
		UNLOCK_SCR ();
	}

	return 0;
}

/* Show menu in the main window after beeing in other mode */
static void return_menu () {
	LOCK_SCR ();

	main_win_mode = WIN_MENU;
	delwin (main_subwin);
	werase (main_win);
	menu_draw (file_menu->menu);
	
	UNLOCK_SCR ();
	
	main_title (old_title);
}

#ifdef VERSION_CHECKER

/* Run a web browser with a parameter */
static void run_browser (char *param) {
	char command[512];
	char *browser;
	pid_t pid;
	char *argv[3];
	int old_mode;

	if ((browser = getenv("BROWSER")))
		strcpy (command, browser);
	else
		strcpy (command, "lynx");
	
	/* Leave ncurses mode */
	LOCK_SCR ();
	def_prog_mode ();
	endwin ();
	old_mode = main_win_mode;
	main_win_mode = WIN_NOCURSES;
	UNLOCK_SCR ();

	/* Run command in a separate process */
	if ((pid = fork()) == 0) {
		argv[0] = command;
		argv[1] = param;
		argv[2] = NULL;
		execvp (command, argv);
		exit (0);
	}
	else
		waitpid (pid, NULL, 0);
	
	/* Back to ncurses mode */
	LOCK_SCR ();
	reset_prog_mode ();
	refresh ();
	main_win_mode = old_mode;
	UNLOCK_SCR ();
	update_info_win ();
}

/* Run a www browser to download the moc */
static void download_moc () {
	char *dir = get_dir ();
	
	chdir (getenv("HOME"));
	run_browser (DOWNLOAD_URL);
	chdir (dir);
}

#endif

/* Handle key events */
void events () {
	int ch;
			
	while (1) {
		ch = wgetch (main_win);
		if (main_win_mode == WIN_MENU) {
			if (menu_command(ch))
				break;
		}
#ifdef VERSION_CHECKER
		else if (main_win_mode == WIN_VER_CHECKER) {
			switch (ch) {
				case 'N':
				case 'n':
					run_browser (NEWS_URL);
					break;
				case 'D':
				case 'd':
					download_moc ();
					break;
				case 'Q':
				case 'q':
				case ' ':
					return_menu ();
					break;
				default:
			}
		}
#endif
		else
			return_menu ();
	}
}

/* Move status line */
void move_progress_bar () {
	int i;
	
	LOCK_SCR ();
	for (i = 2; i >= 0; i--) {

		if (progress_bar.pos[i] == 0)
			progress_bar.direction[i] = 1;
		else if (progress_bar.pos[i] == 9)
			progress_bar.direction[i] = -1;
		progress_bar.pos[i] += progress_bar.direction[i];
	}
	UNLOCK_SCR ();

	update_info_win ();
}
