/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#ifdef HAVE_PRIORITY
# include <sys/resource.h>
#endif
#include <errno.h>
#include <string.h>
#include "getopt.h"
#include "interface.h"
#include "audio.h"
#include "filelist.h"
#include "playlist.h"
#include "config.h"
#include "options.h"
#include "main.h"

/* Allocate memory with checking */
void *cmalloc (size_t size) {
	void *mem;

	if ((mem = malloc (size)) == NULL)
		fatal ("Can't allocate memory!");

	return mem;
}

static void show_ver () {
	printf (PACKAGE_STRING" ");

	/* Show build time */
#ifdef __DATE__
	printf ("Build: "__DATE__" ");
# ifdef __TIME__
	printf (__TIME__);
# endif /* __TIME__ */
#endif /* __DATE__ */

	putchar ('\n');

	/* Show compiled-in components */
	printf ("Compiled with:");
#ifdef HAVE_MAD
	printf (" MP3");
#endif
#ifdef HAVE_VORBIS
	printf (" OGG");
#endif
#ifdef HAVE_OSS
	printf (" mixer");
#endif
#ifdef VERSION_CHECKER
	printf (" version checker");
#endif
	
	putchar ('\n');
	
}

/* Show program usage and exit */
static void usage (const char *prg_name) {
	printf ("\
"PACKAGE_STRING"
Usage:
%s [OPTIONS]... [FILE]...
-V --version		Show program version and exit.
-h --help		Show usage and exit.
-r --recursively	Add files in direstories given as an argument
			recursively.
-f --fullinfo		Show titles (eg from id3tags) instead of names in the
			menu (default).
-s --shortinfo		Show file names in the menu (opposite to -f).
-m --musicdir		Go to your music directory after start (requires such
			entry in the config file).
-e --errors		Show errors in streams.
-n --no-errors		Do not show errors in streams (opposite to -e).
-p --priority N		Set the priority to a given value.
-l --dont-play		Don't start playing files added as command line
			arguments.
-R --repeat		Turn on repeat.
-E --no-repeat		Turn off repeat.
", prg_name);
}

/* Return the current directory name */
char *curr_dir () {
	static char dir[PATH_MAX];

	if (getcwd (dir, sizeof(dir)) == NULL)
		fatal ("getcwd() failed!");

	return dir;
}

/* Add all files and directories gives as an argument.
 * if recurr is >0, add files in directories recursively
 */
static void add_files (char *argv[], const int argc, const int first,
		const short rec, const short play) {
	int i;

	playlist = new_plist ();
	
	for (i = first; i < argc; i++) {
		short type = read_file_type(argv[i]);
		
		if (type == F_DIRECTORY && rec)
			add_dir (argv[i], playlist, 1);
		else if (type != F_DIRECTORY)
			add_file (playlist, argv[i], 0);
	}

	if (playlist->next) {
		file_menu = make_plist_menu (playlist);
		set_num_items (file_menu->menu->nitems);
		if (play) {
			hidden_list = playlist;
			play_it (playlist->next);
		}
		curr_plist = playlist;
		main_title ("PLAYLIST");
	}
	else {
		enter_dir (curr_dir());
		show_err ("No files could be added.");
	}
}

#ifdef HAVE_PRIORITY

/* Set the priority */
static void priority () {
	uid_t uid;

	if (options.priority != 0) {
		uid = getuid ();
		if (setuid(0) == -1)
			return;
		setpriority (PRIO_PROCESS, 0, options.priority);
		setuid (uid);
		seteuid (uid);
	}
}

#endif

int main (int argc, char *argv[]) {
	int ret, opt_index = 0;
	struct option long_options[] = {
		{ "version",		0, NULL, 'V' },
		{ "help",		0, NULL, 'h' },
		{ "recursively", 	0, NULL, 'r' },
		{ "fullinfo",		0, NULL, 'f' },
		{ "shortinfo",		0, NULL, 's' },
		{ "musicdir",		0, NULL, 'm' },
		{ "no-errors",		0, NULL, 'n' },
		{ "errors",		0, NULL, 'e' },
		{ "priority",		1, NULL, 'p' },
		{ "dont-play",		0, NULL, 'l' },
		{ "repeat",		0, NULL, 'R' },
		{ "no-repeat",		0, NULL, 'E' },
		{ 0, 0, 0, 0}
	};

	/* Options */
	short recurr = 0;
	short go_musicdir = 0;
	short start_playing = 1;

	show_err (PACKAGE_STRING", report bugs to "PACKAGE_BUGREPORT);

	read_config ();

	/* Process program arguments */
	while ((ret = getopt_long(argc, argv, "Vhrfsmenp:lRE", long_options,
					&opt_index)) != -1) {
		switch (ret) {
			case 'V':
				show_ver ();
				exit (0);
				break;
			case 'h':
				usage (argv[0]);
				exit (0);
				break;
			case 'r':
				recurr = 1;
				break;
			case 'f':
				options.full_info = 1;
				break;
			case 's':
				options.full_info = 0;
				break;
			case 'm':
				if (options.music_dir)
					go_musicdir = 1;
				break;
			case 'e':
				options.show_errors = 1;
				break;
			case 'n':
				options.show_errors = 0;
				break;
			case 'p':
				options.priority = atoi (optarg);
				break;
			case 'l':
				start_playing = 0;
				break;
			case 'R':
				options.repeat = 1;
				break;
			case 'E':
				options.repeat = 0;
				break;
			default:
				usage (argv[0]);
				exit (1);
		}
	}

#ifdef HAVE_PRIORITY
	priority ();
#endif

	init_audio ();
	init_interface ();

#ifdef VERSION_CHECKER
	if (need_remind())
		show_err ("You can press 'V' to check for a new version of the "
				"MOC!");
#endif

	progress_bar.on = 1;
	update_info_win ();
	if (go_musicdir)
		enter_dir (options.music_dir);
	else if (optind == argc)
		enter_dir (curr_dir());
	else if (optind + 1 == argc
			&& read_file_type(argv[optind]) == F_DIRECTORY
			&& !recurr) {
		if (enter_dir(argv[optind]))
			fatal ("Can't go to this directory.\n");
	}
	else
		add_files (argv, argc, optind, recurr, start_playing);
	LOCK_SCR ();
	progress_bar.on = 0;
	UNLOCK_SCR ();
	update_info_win ();
	
	srand (time(NULL));
	events ();
	stop ();
	end_interface ();

	return 0;
}
