/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <ncurses.h>
#include "main.h"
#include "menu.h"

void menu_draw (Menu *menu) {
	int i, j;

	for (i = menu->top; i < menu->nitems && i - menu->top < menu->maxy;
			i++) {
		wmove (menu->win, i - menu->top + 1, 1);
		
		/* Set attributes */
		if (i == menu->selected)
			wattrset (menu->win, menu->items[i]->attr_sel);
		else
			wattrset (menu->win, menu->items[i]->attr_normal);
		
		waddnstr (menu->win, menu->items[i]->title, menu->maxx);
		
		/* Make blank line to the right side of the screen */
		if (i == menu->selected) {
			for (j = strlen (menu->items[i]->title) + 1;
					j <= menu->maxx; j++)
				waddch (menu->win, ' ');
		}
	}
	wattrset (menu->win, COLOR_PAIR(1));
}

Menu *menu_new (WINDOW *win, MenuItem **items) {
	Menu *menu;
	int i;
	
	menu = (Menu *)malloc (sizeof(Menu));
	if (menu == NULL || items == NULL)
		return NULL;

	menu->win = win;
	menu->items = items;
	menu->top = 0;
	menu->selected = 0;
	getmaxyx (win, menu->maxy, menu->maxx);
	menu->maxx -= 2; /* The window contains border */
	menu->maxy--; /* Border without bottom line */

	/* Count items */
	for (i = 0; items[i] != NULL; i++);
	menu->nitems = i;

	menu_draw (menu);

	return menu;
}

MenuItem *menu_newitem (char *title) {
	MenuItem *item;

	item = (MenuItem *)malloc (sizeof(MenuItem));
	if (item == NULL)
		return NULL;

	item->title = title;
	item->usrptr = NULL;
	item->attr_normal = COLOR_PAIR(1);
	item->attr_sel = COLOR_PAIR(2);

	return item;
}

void menu_free (Menu *menu) {
	int i;

	/* Free items */
	for (i = 0; i < menu->nitems; i++)
		free (menu->items[i]);

	/* Free menu */
	free (menu);
}

void menu_driver (Menu *menu, int req) {
	if (req == REQ_DOWN && menu->selected < menu->nitems - 1) {
		menu->selected++;
		if (menu->selected > menu->top + menu->maxy - 1) {
			menu->top = menu->selected - (menu->maxy - 1) / 2;
			if (menu->top > menu->nitems - menu->maxy)
				menu->top = menu->nitems - menu->maxy;
		}
	}
	else if (req == REQ_UP && menu->selected > 0) {
		menu->selected--;
		if (menu->top > menu->selected) {
			menu ->top = menu->selected - (menu->maxy - 1) / 2;
			if (menu->top < 0)
				menu->top = 0;
		}
	}
	else if (req == REQ_PGDOWN && menu->selected < menu->nitems - 1) {
		if (menu->selected + menu->maxy - 1 < menu->nitems - 1) {
			menu->selected += menu->maxy - 1;
			menu->top += menu->maxy - 1;
			if (menu->top > menu->nitems - menu->maxy)
				menu->top = menu->nitems - menu->maxy;
		}
		else {
			menu->selected = menu->nitems - 1;
			menu->top = menu->nitems - menu->maxy;
		}

		if (menu->top < 0)
			menu->top = 0;

	}
	else if (req == REQ_PGUP && menu->selected > 0) {
		if (menu->selected - menu->maxy + 1 > 0) {
			menu->selected -= menu->maxy - 1;
			menu->top -= menu->maxy - 1;
			if (menu->top < 0)
				menu->top = 0;
		}
		else {
			menu->selected = 0;
			menu->top = 0;
		}
	}
	else if (req == REQ_TOP) {
		menu->selected = 0;
		menu->top = 0;
	}
	else if (req == REQ_BOTTOM) {
		menu->selected = menu->nitems - 1;
		menu->top = menu->selected - menu->maxy + 1;
		if (menu->top < 0)
			menu->top = 0;
	}
}

MenuItem *menu_curritem (Menu *menu) {
	return menu->items[menu->selected];
}

/* Make this item selected */
void menu_setcurritem (Menu *menu, int num) {
	menu->selected = num;

	if (menu->selected < menu->top)
		menu->top = menu->selected;
	else if (menu->selected > menu->top + menu->maxy - 1)
		menu->top = menu->selected - menu->maxy + 1;	

	menu_draw (menu);
}

/* Return number of the item (number in table), return -1 if not found */
int menu_find_item (Menu *menu, const char *title) {
	int i;

	for (i = 0; i < menu->nitems; i++)
		if (!strcmp(menu->items[i]->title, title))
			break;

	/* Not found? */
	if (i == menu->nitems)
		return -1;
	
	return i;
}

void set_menu_state (Menu *menu, int selected, int top) {
	if (selected >= menu->nitems)
		menu->selected = menu->nitems - 1;
	else
		menu->selected = selected;

	if (top > menu->nitems - menu->maxy + 1)
		menu->top = menu->nitems - menu->maxy;
	else
		menu->top = top;

	if (menu->top < 0)
		menu->top = 0;
}
