/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

/* This code was writen od the basis of madlld.c (C) by Bertrand Petit 
 * including code from xmms-mad (C) by Sam Clegg and winamp plugin for madlib
 * (C) by Robert Leslie.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <mad.h>
#include <id3tag.h>
#include "main.h"
#include "audio.h"
#include "interface.h"
#include "playlist.h"
#include "xing.h"
#include "options.h"

#define INPUT_BUFFER	(8 * 1024)

/* Information about the file */
struct {
	int infile; /* fd on the mp3 file */
	unsigned long bitrate;
	unsigned int freq;
	short channels;
	signed long duration; /* Total time of the file in seconds*/
	off_t size; /* Size of the file */
	struct xing xing;
	short has_xing;
} info;

pthread_mutex_t seek_mut = PTHREAD_MUTEX_INITIALIZER;
int seek;

static unsigned char in_buff[INPUT_BUFFER];

/* Fill in the mad buffer, return number of bytes readed, 0 on eof, < 0 on error */
static size_t fill_buff (struct mad_stream *stream) {
	size_t remaining, read_size;
	unsigned char *read_start;
	
	if (stream->next_frame != NULL) {
		remaining = stream->bufend - stream->next_frame;
		memmove (in_buff, stream->next_frame, remaining);
		read_start = in_buff + remaining;
		read_size = INPUT_BUFFER - remaining;
	}
	else {
		read_start = in_buff;
		read_size = INPUT_BUFFER;
		remaining = 0;
	}

	read_size = read (info.infile, read_start, read_size);
	if (read_size < 0) {
		show_err ("read() failed: %s\n", strerror (errno));
		return -1;
	}
	else if (read_size == 0)
		return 0;

	mad_stream_buffer(stream, in_buff, read_size + remaining);
	stream->error = 0;

	return read_size;
}

char *get_tag (struct id3_tag *tag, const char *what) {
	struct id3_frame *frame;
	union id3_field *field;
	const id3_ucs4_t *ucs4;
	char *comm = NULL;

	frame = id3_tag_findframe (tag, what, 0);
	if (frame && (field = &frame->fields[1])) {
		ucs4 = id3_field_getstrings (field, 0);
		if (ucs4)
			comm = id3_ucs4_latin1duplicate (ucs4);
	}

	return comm;
}

/* Fill info structure with data from the id3 tag */
void read_mp3_info (const char *file_name, song_info *info) {
	struct id3_tag *tag;
	struct id3_file *id3file;
	char *artist = NULL, *title = NULL, *album = NULL;

	/* read the ID3 tag */
	id3file = id3_file_open (file_name, ID3_FILE_MODE_READONLY);
	if (!id3file)
		return;
	tag = id3_file_tag (id3file);
	if (tag) {
		artist = get_tag (tag, ID3_FRAME_ARTIST);
		title = get_tag (tag, ID3_FRAME_TITLE);
		album = get_tag (tag, ID3_FRAME_ALBUM);
	}
	id3_file_close (id3file);

	info->title = title;
	info->album = album;
	info->artist = artist;
}

int load_mp3 (char *file_name) {
	struct stat stat;

	/* Reset information about the file */
	info.bitrate = 0;
	info.freq = 0;
	info.channels = 0;
	info.duration = 0;
	info.has_xing = 0;

	/* Open the file */
	if ((info.infile = open (file_name, O_RDONLY)) == -1) {
		show_err ("open() failed: %s\n", strerror (errno));
		return 1;
	}

	if (fstat(info.infile, &stat) == -1) {
		show_err ("Can't stat() file: %s\n", strerror(errno));
		return 1;
	}

	info.size = stat.st_size;

	seek = 0;

	return 0;
}

/* Scale PCM data to 16 bit unsigned */
static inline signed int scale (mad_fixed_t sample) {
	
	/* round */
	sample += (1L << (MAD_F_FRACBITS - 16));

	/* clip */
	if (sample >= MAD_F_ONE)
		sample = MAD_F_ONE - 1;
	else if (sample < -MAD_F_ONE)
		sample = -MAD_F_ONE;

	/* quantize */
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}

static void write_output (struct mad_pcm *pcm, struct mad_header *header) {
	unsigned int nsamples;
	mad_fixed_t const *left_ch, *right_ch;
	char *output;
	int olen = 0;
	int pos = 0;

	nsamples = pcm->length;
	left_ch = pcm->samples[0];
	right_ch = pcm->samples[1];
	olen = nsamples * MAD_NCHANNELS (header) * 2;
	output = (char *) cmalloc (olen * sizeof (char));

	while (nsamples--) {
		signed int sample;

		/* output sample(s) in 16-bit signed little-endian PCM */
		sample = scale (*left_ch++);
		output[pos++] = (sample >> 0) & 0xff;
		output[pos++] = (sample >> 8) & 0xff;

		if (MAD_NCHANNELS (header) == 2) {
			sample = scale (*right_ch++);
			output[pos++] = (sample >> 0) & 0xff;
			output[pos++] = (sample >> 8) & 0xff;
		}
	}
	play_buff (output, olen);
	free (output);
}

void process_mp3 () {
	short first_frame = 1;
	short stop;
	int do_seek, seek_skip = 0;
	struct mad_stream stream;
	struct mad_frame frame;
	struct mad_synth synth;
	mad_timer_t timer;

	mad_stream_init (&stream);
	mad_frame_init (&frame);
	mad_synth_init (&synth);
	mad_timer_reset (&timer);

	while (1) {

		/* Fill the input buffer */
		if (stream.buffer == NULL || stream.error == MAD_ERROR_BUFLEN) {
			if (fill_buff(&stream) <= 0)
				break;
		}
		
		pthread_mutex_lock (&seek_mut);
		do_seek = seek;
		pthread_mutex_unlock (&seek_mut);

		/* Skip 2 frames after seeking */
		if (seek_skip) {
			int skip = 2;

			do {
				if (mad_frame_decode(&frame, &stream) == 0) {
		  			mad_timer_add(&timer,
							frame.header.duration);
					if (--skip == 0)
						mad_synth_frame(&synth, &frame);
				}
				else if (!MAD_RECOVERABLE(stream.error))
					break;
			} while (skip);

			seek_skip = 0;
		}
		
		if (seek != 0) {
			int new_position;

			seek += mad_timer_count (timer, MAD_UNITS_SECONDS);

			if (seek >= info.duration)
				seek = info.duration;
			else if (seek < 0)
				seek = 0;

			new_position = ((double) seek /
					(double) info.duration) * info.size;
			if (lseek (info.infile, new_position, SEEK_SET) == -1)
				show_err ("Failed to seek to: %d", new_position);
			else {
				mad_timer_set (&timer, seek, 0, 0);

				mad_frame_mute (&frame);
				mad_synth_mute (&synth);
				stream.error = MAD_ERROR_BUFLEN;
				stream.sync = 0;
				stream.next_frame = NULL;
				seek = 0;
				seek_skip = 1;
			}

			pthread_mutex_lock (&seek_mut);
			do_seek = 0;
			pthread_mutex_unlock (&seek_mut);

			continue;
		}

		if (mad_frame_decode (&frame, &stream)) {
			if (MAD_RECOVERABLE(stream.error)) {

				/* Ignore LOSTSYNC */
				if (stream.error == MAD_ERROR_LOSTSYNC)
					continue;

				if (options.show_errors)
					show_err ("Broken frame: %s",
							mad_stream_errorstr (&stream));
				continue;
			}
			else if (stream.error == MAD_ERROR_BUFLEN)
				continue;
			else {
				if (options.show_errors)
					show_err ("Broken frame: %s",
							mad_stream_errorstr (&stream));
				break;
			}
		}

		mad_timer_add (&timer, frame.header.duration);

		/* Bitrate and number or channels can change */
		if (info.channels != MAD_NCHANNELS(&frame.header) ||
				info.freq != frame.header.samplerate) {
			if (!first_frame)
				close_audio ();
			
			
			if ((info.freq = frame.header.samplerate) == 0) {
				if (options.show_errors)
					show_err ("Broken file: information "
							"about the frequency "
							"couldn't be readed.");
				break;
			}
			info.channels = MAD_NCHANNELS(&frame.header);
			
			if (open_audio (16, info.channels, info.freq))
				break;

			set_stereo (info.channels == 2 ? 1 : 0);
			set_rate (info.freq / 1000);
		}

		/* Change of the bitrate? */
		if (frame.header.bitrate != info.bitrate) {
			if ((info.bitrate = frame.header.bitrate) == 0) {
				show_err ("Broken file: information "
						"about the bitrate couldn't "
						"be readed.\n");
				break;
			}

			set_bitrate (info.bitrate / 1000);
		}

		/* Read some information about the strean from first frame */
		if (first_frame) {
			mad_timer_t duration = timer;

			if (xing_parse(&info.xing, stream.anc_ptr,
						stream.anc_bitlen) != -1) {
				xing_init (&info.xing);
				mad_timer_multiply (&duration,
						info.xing.frames);
				info.has_xing = 1;
			}
			else {
				mad_timer_multiply (&duration, info.size /
					(stream.next_frame - stream.this_frame));
			}

			info.duration = mad_timer_count (duration,
					MAD_UNITS_SECONDS);
			set_time (info.duration);
			
			first_frame = 0;
		}
		
		set_curr_time (mad_timer_count (timer, MAD_UNITS_SECONDS));

		mad_synth_frame (&synth, &frame);
		mad_stream_sync (&stream);
		write_output (&synth.pcm, &frame.header);

		/* Pause? */
		pthread_mutex_lock (&paused_mut);
		if (paused)
			pthread_cond_wait (&paused_cond, &paused_mut);
		pthread_mutex_unlock (&paused_mut);

		/* Stop ? */
		pthread_mutex_lock (&req_mut);
		stop = stop_req;
		pthread_mutex_unlock (&req_mut);
		if (stop)
			break;
	}

	mad_stream_finish (&stream);
	mad_frame_finish (&frame);
	mad_synth_finish (&synth);

	close (info.infile);
	close_audio ();
	reset_file_info ();
}

void seek_mp3 (int sec) {
	pthread_mutex_lock (&seek_mut);
	seek = sec;
	pthread_mutex_unlock (&seek_mut);
}
