/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdio.h> /* Only for debug! */
#include <errno.h>
#include <pthread.h>
#include <vorbis/vorbisfile.h>
#include <vorbis/codec.h>
#include "audio.h"
#include "interface.h"
#include "playlist.h"
#include "options.h"

#define PCM_BUFF_SIZE	4096

static FILE *file;
static OggVorbis_File vf;
static pthread_mutex_t vf_mut = PTHREAD_MUTEX_INITIALIZER;
static int cancel_state;
static char pcm_buff[PCM_BUFF_SIZE];

#define LOCK_VF() pthread_mutex_lock (&vf_mut); \
	pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, &cancel_state)
#define UNLOCK_VF() pthread_mutex_unlock (&vf_mut); \
	pthread_setcancelstate (cancel_state, NULL)

/* Fill info structure with data from ogg comments */
void read_ogg_info (const char *file_name, song_info *info) {
	int i;
	vorbis_comment *comments;
	OggVorbis_File vf;
	char *title = NULL, *artist = NULL, *album = NULL;

	if (!(file = fopen (file_name, "r"))) {
		show_err ("Can't load OGG: %s", strerror(errno));
		return;
	}

	if (ov_test(file, &vf, NULL, 0) < 0) {
		show_err ("ov_test() failed!");
		return;
	}

	comments = ov_comment (&vf, -1);
	for (i = 0; i < comments->comments; i++) {
		if (!strncasecmp(comments->user_comments[i], "title=",
					strlen ("title=")))
			title = comments->user_comments[i] + strlen ("title=");
		else if (!strncasecmp(comments->user_comments[i], "artist=",
					strlen ("artist=")))
			artist = comments->user_comments[i] + strlen ("artiat=");
		else if (!strncasecmp(comments->user_comments[i], "album=",
					strlen ("album=")))
			album = comments->user_comments[i] + strlen ("album=");
	}

	if (title)
		info->title = strdup (title);
	else
		info->title = NULL;
	
	if (artist)
		info->artist = strdup (artist);
	else
		info->artist = NULL;
	
	if (album)
		info->album = strdup (album);
	else
		info->album = NULL;

	ov_clear (&vf);
}

/* Load .ogg file - init vf structure, return 1 on failture */
int load_ogg (char *file_name) {
	vorbis_info *info;

	if (!(file = fopen (file_name, "r"))) {
		show_err ("Can't load OGG: %s", strerror(errno));
		return 1;
	}

	if (ov_open(file, &vf, NULL, 0) < 0) {
		show_err ("ov_open() failed!");
		return 1;
	}

	info = ov_info (&vf, -1);
	if (open_audio (16, info->channels, info->rate))
		return 1;
	
	set_rate (info->rate);
	set_time (ov_time_total (&vf, -1));
	if (info->channels == 2)
		set_stereo (1);
	else
		set_stereo (0);

	return 0;
}

/* Decode and play stream */
void process_ogg () {
	int current_section, eof = 0;
	int last_section = -1;
	long ret, bitrate;
	short stop = 0;
	vorbis_info *info;

	while (!eof && !stop) {
		pthread_mutex_lock (&req_mut);
		stop = stop_req;
		pthread_mutex_unlock (&req_mut);
		LOCK_VF ();
		ret = ov_read(&vf, pcm_buff, PCM_BUFF_SIZE, 0,
				2, 1, &current_section);
		UNLOCK_VF ();
		if (ret ==0)
			eof = 1;
		else if (ret < 0) {
			if (options.show_errors)
				show_err ("Error in the stream!");
		}
		else {
			if (last_section != current_section
					&& last_section >= 0) {
				close_audio ();
				info = ov_info (&vf, -1);
				if (open_audio (16, info->channels, info->rate))
					return;
			}
			play_buff (pcm_buff, ret);
		}

		/* Pause? */
		pthread_mutex_lock (&paused_mut);
		if (paused)
			pthread_cond_wait (&paused_cond, &paused_mut);
		pthread_mutex_unlock (&paused_mut);

		/* Update the current time and the bitrate */
		LOCK_VF ();
		bitrate = ov_bitrate_instant (&vf);
		if (bitrate > 0)
			set_bitrate (bitrate / 1000);
		set_curr_time (ov_time_tell (&vf));
		UNLOCK_VF ();
		last_section = current_section;
	}
	
	ov_clear (&vf);
	close_audio ();

	reset_file_info ();
}

void seek_ogg (int ms) {
	LOCK_VF ();
	ov_time_seek (&vf, ov_time_tell (&vf) + (double)ms);
	UNLOCK_VF ();
}
