/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include "options.h"
#include "main.h"
#include "interface.h"

#define CONFIG_DIR	".moc"
#define REMINDER_FILE	"reminder_time"
#define CONFIG_FILE	"config"

#define CONFIG_NAME_LEN		30
#define CONFIG_VALUE_LEN	256

/* isblank() is a GNU extension */
#ifndef isblank
#define isblank(c) (c == ' ' || c == '\t')
#endif

struct options options = {
	1,	/* full_info */
	NULL,	/* music_dir */
	0,	/* show_errors */
	31 * 24 * 60 * 60,	/* reminder (one month) */
	0,	/* Priority */
	0,	/* Shuffle */
	0	/* Repeat */
};

pthread_mutex_t options_mut = PTHREAD_MUTEX_INITIALIZER;

/* Use parameter from config file */
static int exec_param (const char *name, const char *value) {
	if (!strcasecmp(name, "FullInfo")) {
		if (!strcasecmp(value, "yes"))
			options.full_info = 1;
		else
			options.full_info = 0;
	}
	else if (!strcasecmp(name, "MusicDir")) {
		options.music_dir = strdup (value);
	}
	else if (!strcasecmp(name, "ShowStreamErrors")) {
		if (!strcasecmp(value, "yes"))
			options.show_errors = 1;
		else
			options.show_errors = 0;
	}
	else if (!strcasecmp(name, "Reminder")) {
		options.remind = atoi (value) * 24 * 60 * 60;
	}
	else if (!strcasecmp(name, "Priority")) {
		options.priority = atoi (value);
	}
	else if (!strcasecmp(name, "Repeat")) {
		if (!strcasecmp(value, "yes"))
			options.repeat = 1;
		else
			options.repeat = 0;
	}
	else if (!strcasecmp(name, "Shuffle")) {
		if (!strcasecmp(value, "yes"))
			options.shuffle = 1;
		else
			options.shuffle = 0;
	}
	else
		return 1;
		
	return 0;
}

/* Parse config file */
static void parse_config (FILE *file) {
	int ch;
	short comm = 0, eq = 0, quote = 0, esc = 0;
	char opt_name[30];
	char opt_value[100];
	int line = 1;
	int name_pos = 0;
	int value_pos = 0;

	while ((ch = getc(file)) != EOF) {

		/* Skip comment */
		if (comm && ch != '\n')
			continue;
		
		/* Interpret parameter */
		if (ch == '\n') {
			comm = 0;

			opt_name[name_pos] = 0;
			opt_value[value_pos] = 0;

			if (opt_name[0] != 0
					&& exec_param(opt_name, opt_value)) {
				show_err ("Error in config file, line %d.",
						line);
				return;
			}

			opt_name[0] = 0;
			opt_value[0] = 0;
			name_pos = 0;
			value_pos = 0;
			eq = 0;
			quote = 0;
			esc = 0;

			line++;
		}

		/* Turn on comment */
		else if (ch == '#' && !quote)
			comm = 1;

		/* Turn on quote */
		else if ((ch == '"' || ch == '\'') && !esc)
			quote = !quote;

		else if (ch == '=' && !quote) {
			if (eq) {
				show_err ("Error in config file, line %d.",
						line);
				return;
			}
			
			eq = 1;
			opt_value[0] = 0;
		}

		/* Turn on escape */
		else if (ch == '\\' && !esc) {
			esc = 1;
			continue;
		}

		/* Add char to parameter value */
		else if ((!isblank(ch) || quote) && eq) {
			if (CONFIG_VALUE_LEN == value_pos + 1) {
				show_err ("Error in config file, line %d is "
						"too long.", line);
				return;
			}
			opt_value[value_pos++] = ch;
		}

		/* Add char to parameter name */
		else if (!isblank(ch) || quote) {
			if (CONFIG_NAME_LEN == name_pos + 1) {
				show_err ("Error in config file, line %d is "
						"too long.", line);
				return;
			}
			opt_name[name_pos++] = ch;
		}

		esc = 0;
	}
}

void read_config () {
	char file_name[PATH_MAX];
	struct stat test_stat;
	FILE *config_file;

	/* Check if our config directory exists */
	sprintf (file_name, "%s/"CONFIG_DIR, getenv("HOME"));
	if (stat(file_name, &test_stat) == -1) {

		/* No such file? Create it and exit, because there is
		 * no config file */
		if (errno == ENOENT) {
			if (mkdir(file_name, 0700) != -1)
				show_err ("Creating config directory: %s",
						file_name);
			else
				show_err ("Cant create config directory: %d.",
						strerror(errno));
			
		}
		else
			show_err ("Can't access config directory: %s.",
					strerror(errno));
		
		return;	
	}
	
	sprintf (file_name, "%s/"CONFIG_DIR"/"CONFIG_FILE, getenv("HOME"));
	if ((config_file = fopen (file_name, "r")) == NULL) {
		if (errno != ENOENT)
			show_err ("Can't read config file: %s.",
					strerror(errno));
		return;
	}

	parse_config (config_file);
	fclose (config_file);
}

#ifdef VERSION_CHECKER

/* Make file with the current time */
static void make_remind_file (const char *file_name) {
	FILE *file;

	if ((file = fopen(file_name, "w")) == NULL) {
		show_err ("Can't create reminder file: %s.",
					strerror(errno));
		return;
	}

	fprintf (file, "%ld", time(NULL));

	fclose (file);
}

/* Return 1 if we must remind of version checker */
int need_remind () {
	FILE *remind_file;
	char file_name[PATH_MAX];
	time_t last_time, curr_time;

	if (options.remind == 0)
		return 0;

	sprintf (file_name, "%s/"CONFIG_DIR"/"REMINDER_FILE, getenv("HOME"));
	if ((remind_file = fopen (file_name, "r")) == NULL) {
		if (errno != ENOENT) {
			show_err ("Can't read reminder file: %s.",
					strerror(errno));
			return 0;
		}
		else {
			make_remind_file (file_name);
			return 1;
		}
	}
	
	fscanf (remind_file, "%ld", &last_time);
	curr_time = time (NULL);
	fclose (remind_file);

	if (curr_time - options.remind > last_time) {
		make_remind_file (file_name);
		return 1;
	}
	else
		return 0;
}

#endif
