/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <ncurses.h>
#include <pthread.h>
#include "main.h"
#include "playlist.h"
#include "interface.h"
#include "audio.h"

plist *playlist = NULL;
plist *tmp_list = NULL;
plist *hidden_list = NULL;
shuffled_list *shuffled = NULL;

pthread_mutex_t hidden_list_mut = PTHREAD_MUTEX_INITIALIZER;

void free_song_info (song_info *info) {
	if (info) {
		if (info->title)
			free (info->title);
		if (info->album)
			free (info->album);
		if (info->artist)
			free (info->artist);
	}
}

static void free_item (plist *item) {
	if (item->item.file_name)
		free (item->item.file_name);
	free_song_info (item->item.info);
	if (item->item.title)
		free (item->item.title);
	free (item);
}

/* Make new playlist */
plist *new_plist () {
	plist *item;
	
	/* Fake element */
	item = (plist *)cmalloc (sizeof(plist));
	item->item.file_name = NULL;
	item->item.title = NULL;
	item->item.info = NULL;
	item->item.menu_item = NULL;
	item->next = NULL;

	return item;
}

/* Free playlist */
void free_plist (plist *list) {
	plist *citem, *pitem;
	
	citem = list;
	while (citem) {
		pitem = citem;
		citem = citem->next;
		free_item (pitem);
	}
}

/* Duplicate pitem */
plist *dup_item (plist *item) {
	plist *new_item;

	new_item = (plist *)cmalloc (sizeof(plist));
	
	new_item->item.file_name = strdup (item->item.file_name);
	new_item->item.title = strdup (item->item.title);
	new_item->item.type = item->item.type;

	if (item->item.info) {
		new_item->item.info = (song_info *)cmalloc (sizeof(song_info));
		new_item->item.info->title = item->item.info->title ?
			strdup (item->item.info->title) : NULL;
		new_item->item.info->artist = item->item.info->artist ?
			strdup (item->item.info->artist) : NULL;
		new_item->item.info->album = item->item.info->album ?
			strdup (item->item.info->album) : NULL;
	}
	else
		new_item->item.info = NULL;

	return new_item;
}

/* Duplicate plist, skip directories */
plist *dup_plist (plist *list) {
	plist *new_item, *citem, *new_list;

	new_item = new_plist ();
	new_list = new_item;
	for (citem = list->next; citem != NULL; citem = citem->next)
		if (citem->item.type != F_DIRECTORY) {
			new_item->next = dup_item (citem);
			new_item = new_item->next;
		}
	
	new_item->next = NULL;
	return new_list;
}

void add_plist_item (plist *list, plist *item, const short sort) {
	plist *clist;
	char *title = item->item.title;
	short type = item->item.type;
		
	/* Go before the greater element than this file, directories
	 * on the top
	 */
	for (clist = list; clist->next != NULL;
			clist = clist->next) {
		
		if (!sort)
			continue;

		/* FIXME: can it be simpler? */
		if ((strcmp(clist->next->item.title, title) > 0
				&& ((clist->next->item.type == F_DIRECTORY
					&& type == F_DIRECTORY)
				|| (clist->next->item.type != F_DIRECTORY
					&& type != F_DIRECTORY)))
				|| (clist->next->item.type != F_DIRECTORY
					&& type == F_DIRECTORY))
			break;
	}

	/* Add the file between cfile and cfile->next */
	item->next = clist->next;
	clist->next = item;
}

/* Add new file to the plat list */
void add_to_plist (plist *list, const char *file_name, const char *title,
		song_info *info, const short type, const short sort) {
	plist *new_item;

	new_item = (plist *)cmalloc (sizeof(plist));
	new_item->item.file_name = strdup (file_name);
	new_item->item.title = strdup (title);
	new_item->item.type = type;
	new_item->item.info = info;

	add_plist_item (list, new_item, sort);
}

/* Append the new_list at the end of the list */
void append_plist (plist *list, plist *new_list) {
	plist *clist;

	/* new_list can't be empty */
	if (!new_list || !new_list->next)
		return;
	
	/* Go to the end of list */
	for (clist = list; clist->next != NULL;	clist = clist->next);

	clist->next = new_list->next;
	free_item (new_list);
}

/* Make menu from the playlist */
plist_menu *make_plist_menu (plist *list) {
	plist_menu *menu;
	plist *citem;
	int nitems = 0;
	int i = 0;

	menu = (plist_menu *)cmalloc (sizeof(plist_menu));

	for (citem = list->next; citem != NULL; citem = citem->next) {
		citem->item.menu_item = menu_newitem (citem->item.title);
		citem->item.menu_item->usrptr = (void *)citem;
		if (citem->item.type == F_DIRECTORY) {
			citem->item.menu_item->attr_normal =
				COLOR_PAIR(1) | A_BOLD;
			citem->item.menu_item->attr_sel = 
				COLOR_PAIR(2) | A_BOLD;
		}
		
		nitems++;
	}

	menu->items = (MenuItem **)cmalloc (sizeof(MenuItem) * (nitems + 1));
	for (citem = list->next; citem != NULL; citem = citem->next)
		menu->items[i++] = citem->item.menu_item;
	menu->items[nitems] = NULL;
	
	menu->list = list;

	menu->menu = menu_new(main_win, menu->items);

	return menu;
}

void free_plist_menu (plist_menu *menu) {
	menu_free (menu->menu);

	free (menu->items);
	free (menu);
}

void delete_plist_item (plist *list, plist *item) {
	plist *clist;
	
	for (clist = list; clist->next != NULL;
			clist = clist->next)
		if (clist->next == item) {
			plist *to_del = clist->next;
			
			clist->next = to_del->next;
			free_item (to_del);
			break;
		}
}

struct items_table {
	plist *item;
	int num;
};

/* Comparision function for the system qsort() */
static int comp_item (const void *first, const void *second) {
	if (((struct items_table *)first)->num >
			((struct items_table *)second)->num)
		return 1;
	else if (((struct items_table *)first)->num <
			((struct items_table *)second)->num)
		return -1;
	else return 0;
}

/* Shuffle the list, place first_item at the beginning */
shuffled_list *gen_shuffled_list (plist *list, plist* first_item) {
	int nitems = 0, i = 0;
	struct items_table *items_table;
	plist *citem;
	shuffled_list *shuffled, *sh_item;

	/* Make table of pointers to the items */
	for (citem = list->next; citem != NULL; citem = citem->next)
		if (citem != first_item)
			nitems++;
	items_table = (struct items_table *)cmalloc
		(sizeof(struct items_table) * nitems);

	/* Fill the table and assign each entry a random number */
	for (citem = list->next; citem != NULL; citem = citem->next) {
		if (citem == first_item)
			continue;
		items_table[i].item = citem;
		items_table[i++].num = (rand() / (double)RAND_MAX) *
			(double)nitems * (double)10;
	}

	qsort (items_table, nitems, sizeof(struct items_table), comp_item);

	sh_item = (shuffled_list *)cmalloc (sizeof(shuffled_list));
	shuffled = sh_item;
	sh_item->item = first_item;
	for (i = 0; i < nitems; i++) {
		sh_item->next = (shuffled_list *)cmalloc
			(sizeof(shuffled_list));
		sh_item = sh_item->next;
		sh_item->item = items_table[i].item;
	}
	sh_item->next = NULL;
	
	return shuffled;
}

/* Find an item on the shuffled list */
shuffled_list *find_shuffled (shuffled_list *list, const plist *item) {
	shuffled_list *citem;

	for (citem = list; citem != NULL; citem = citem->next)
		if (citem->item == item)
			return citem;

	return NULL;
}

void free_shuffled_list (shuffled_list *list) {
	shuffled_list *citem, *pitem;

	citem = list;
	while (citem) {
		 pitem = citem;
		 citem = citem->next;
		 free (pitem);
	}
}
