/*
 * MOC - music on console
 * Copyright (C) 2002 Damian Pietras <daper@daper.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include "main.h"
#include "interface.h"
#include "version_checker.h"

/* WARNING: This code is not 100% RFC2616 compatible, it is designed to be
 * simple and small, but should works ok.
 * It supports: HTTP 1.0, HTTP 1.1, chunked data.
 */

#ifndef isblank
# define isblank(c) (c == ' ' || c == '\t')
#endif

struct url_data {
	char *host;
	int port;
	char *path;
};

static struct {
	short chunked;
	short status;
	char *err_msg;
	char buff[512];
	int buff_pos;
	int buff_len;
} http_conn;

static int sock;
static int counter;

/* Put string into main window. */
static void put_str (char *format, ...) {
	va_list va;

	va_start (va, format);

	LOCK_SCR ();
	vwprintw (main_subwin, format, va);
	wrefresh (main_subwin);
	UNLOCK_SCR ();

	va_end (va);
}

/* Connect to MOC server */
static int connect_to (struct url_data *url) {
	struct sockaddr_in dest;
	struct hostent *hostname;
	sock = -1;

	put_str ("Resolving %s...\n", url->host);
	
	if (!(hostname = gethostbyname(url->host))) {
		put_str ("%s\n", hstrerror (h_errno));
		return 1;
	}
	
	put_str ("Connecting to %s:%d (%s)...\n", hostname->h_name,
			url->port,
			inet_ntoa(*((struct in_addr *)hostname->h_addr)));

	sock = socket(AF_INET, SOCK_STREAM, 0);
	
	dest.sin_family = AF_INET;
	dest.sin_port = htons(url->port);
	dest.sin_addr = *((struct in_addr *)hostname->h_addr);
	memset(dest.sin_zero, 0, 8);

	if (connect(sock, (struct sockaddr *)&dest, sizeof(struct sockaddr))) {
		put_str ("Can't connect: %s\n", hstrerror (h_errno));
		return 1;
	}

	return 0;
}

static void send_to (const char *buff) {
	if (send(sock, buff, strlen(buff), 0) == -1) {
		perror("Can't send()");
		exit(1);
	}
}

/* Find the position of lf or return -1 */
static int find_lf (const char *buff, int size) {
	int i;
	
	for (i = 0; i < size; i++)
		if (buff[i] == '\n')
			return i;
	
	return -1;
}

/* FIll the buffor, return number of bytes readed */
static int fill_buff (char *buff, int pos, int size) {
	int readed;
	
	if (pos != 0) {
		memmove (buff, buff + pos, size - pos);
		pos = size - pos;
	}

	readed = recv (sock, buff + pos, size - pos, 0);
	if (readed == -1) {
		put_str ("Can't recv().\n");
		return -1;
	}

	return readed;
}

/* Read one line, memory is malloced */
static char *get_line () {
	char *line;
	int readed;
	int lf, cr;

	lf = find_lf (http_conn.buff + http_conn.buff_pos,
			http_conn.buff_len - http_conn.buff_pos);

	if (lf == -1) {
		if ((readed = fill_buff (http_conn.buff, http_conn.buff_pos,
						sizeof(http_conn.buff))) <= 0)
			return NULL;
		
		if (http_conn.buff_pos != 0)
			http_conn.buff_len = sizeof(http_conn.buff) -
				http_conn.buff_pos + readed;
		else
			http_conn.buff_len = readed;

		http_conn.buff_pos = 0;
		
		lf = find_lf (http_conn.buff, http_conn.buff_len);
		if (lf == -1) {
			put_str ("Line too long.\n");
			return NULL;
		}
	}

	cr = 0;
	if (lf > 0 && http_conn.buff[http_conn.buff_pos + lf-1] == '\r')
		cr = 1;
	
	line = (char *)cmalloc (sizeof(char) * (lf + 1 - cr));
	strncpy (line, http_conn.buff + http_conn.buff_pos, lf - cr);
	line[lf-cr] = 0;
	http_conn.buff_pos += lf + 1;

	if (http_conn.chunked) {
		if (counter == 0) {
			counter = strtol (line, NULL, 16);
			if (counter == 0)
				return NULL;
			return get_line ();
		}
		else
			counter -= lf + 1;
	}

	return line;
}

/* Read position of the value of the response header, or 0 if the header
 * is broken.
 */
static int get_value (char *header) {
	int i;
	int coln = 0;
	int len = strlen(header);
	
	for (i = 0; i < len; i++) {
		if (coln && !isblank(header[i]))
			return i;
		if (header[i] == ':')
			coln = 1;
	}

	return 0;
}

/* Read response headers, return != 0 on error */
static int get_headers () {
	char *line;
	int len;
	short chunked = 0;

	http_conn.chunked = 0;

	/* Check HTTP status line */
	line = get_line ();
	len = strlen(line);
	if (len < 14) {
		put_str ("HTTP response header too short.\n");
		return 1;
	}
	if (strncmp(line, "HTTP/1.", sizeof ("HTTP/1.") - 1)
			|| (line[7] != '0' && line[7] != '1')) {
		put_str ("Bad HTTP response.\n");
		return 1;
	}
	
	/* Read HTTP status code and message */
	http_conn.status = atoi (line + sizeof("HTTP/1.X ") - 1);
	http_conn.err_msg = strdup (line + sizeof("HTTP/1.X XXX ") - 1);

	while ((line = get_line()) != NULL && line[0] != 0) {
		int value = get_value (line);

		if (value == 0) {
			put_str ("Broken response header.\n");
			free (line);
			return 1;
		}

		/* Transfer-coding */
		if (!strncasecmp(line, "Transfer-Encoding:",
					sizeof("Transfer-Encoding:")-1)) {
			if (!strncasecmp(line + value, "chunked",
						sizeof("chunked")-1))
				chunked = 1;
			else {
				put_str ("Unknown Transfer-Encoding header.\n");
				free (line);
				return 1;
			}
		}

		free (line);
	}

	if (chunked) {
		http_conn.chunked = 1;
		counter = 0;
	}

	if (line == NULL) {
		put_str ("No response body.\n");
		return 1;
	}

	free (line);

	/* Any HTTP errors? */
	if (http_conn.status != 200) {
		put_str ("HTTP error %d: %s\n", http_conn.status,
				http_conn.err_msg);
		return 1;
	}

	return 0;
}

/* Send the request, return != 0 on error */
static int make_request (const struct url_data *url) {
	char buff[512];

	if (strlen(url->path) > sizeof(buff)) {
		put_str ("Path in the URL is too long.\n");
		return 1;
	}
	if (strlen(url->host) > sizeof(buff)) {
		put_str ("Host in the URL is too long.\n");
		return 1;
	}

	/* Send HTTP request */
	put_str ("Sending request...\n");
	sprintf (buff, "GET %s HTTP/1.1\r\n", url->path);
	send_to (buff);
	sprintf (buff, "Host: %s\r\n", url->host);
	send_to (buff);
	send_to ("Connection: close\n");
	send_to ("User-agent: MOC-version_checker\r\n");
	send_to ("\r\n");

	if (get_headers())
		return 1;

	return 0;
}

/* parse url string */
static struct url_data *parse_url (const char *url) {
	struct url_data *url_data;
	int pos, len;
	
	len = strlen(url);
	
	/* The url can't be shorter than "http://#" */
	if (len < 8)
		return NULL;
	if (strncasecmp(url, "http://", 7))
		return NULL;

	url_data = (struct url_data *)cmalloc (sizeof(struct url_data));
	url_data->port = 80;

	/* Go to the end of host name */
	for (pos = 7; pos < len; pos++)
		if (url[pos] == ':' || url[pos] == '/')
			break;

	/* Copy the host name */
	url_data->host = (char *)cmalloc (sizeof(char) * (pos - 6));
	strncpy (url_data->host, url + 7, pos - 7);
	url_data->host[pos - 7] = 0;

	/* Get port if avaliable */
	if (url[pos] == ':') {
		int port_begin = pos + 1;
		for (pos = port_begin; pos < len; pos++)
			if (!isdigit(url[pos]))
				break;
		
		if (port_begin != pos)
			url_data->port = atoi (url + port_begin);
	}

	/* Get path if avaliable */
	if (len > pos) {
		url_data->path = (char *)cmalloc (sizeof(char) * (len - pos + 1));
		strcpy (url_data->path, url + pos);
	}
	else {
		url_data->path = (char *)cmalloc (sizeof(char) * 2);
		url_data->path[0] = '/';
		url_data->path[1] = 0;
	}
	
	return url_data;
}

/* Return malloced string contained current version of the MOC or NULL on
 * error.
 */
static char *get_curr_version (const struct url_data *url) {
	char *lines[2];
	int i;

	if (make_request(url))
		return NULL;
	
	/* Read response body */
	put_str ("Receiving data...\n");
	for (i = 0; i < 2; i++)
		if ((lines[i] = get_line ()) == NULL) {
			put_str ("Too few response lines.\n");
			return NULL;
		}

	/* Status line */
	if (!strcmp(lines[0], "ERROR")) {
		put_str ("Error: %s\n", lines[1]);
		return NULL;
	}
	else if (strcmp(lines[0], "OK")) {
		put_str ("Error: bad news server response.\n");
		return NULL;
	}

	return lines[1];
	
	free (lines[0]);
}

/* For description see above function */
char *ret_curr_ver () {
	struct url_data *url;
	char *ver = NULL;
	
	url = parse_url (VERSION_URL);
	if (url == NULL) {
		put_str ("Error in the url.\n");
		return NULL;
	}
	else {
		http_conn.buff_len = 0;
		http_conn.buff_pos = 0;
		
		put_str ("Connecting...\n");
		if (connect_to (url))
			return NULL;
		put_str ("Getting information about current version:\n");
		ver = get_curr_version (url);
	}

	if (sock != -1)
		close (sock);

	return ver;
}

void check_version () {
	char *ver;
	
	if (!strcmp(PACKAGE_VERSION, "CVS")) {
		put_str ("Can't check for newer version, you have CVS "
				"version.\n");
		put_str ("Press space.");
	}
	else {
		ver = ret_curr_ver ();
		if (ver == NULL) {
			put_str ("Press space.");
			return;
		}

		if (strcmp(ver, PACKAGE_VERSION)) {
			put_str ("\n   A new version (%s) of the MOC is "
					"available!\n\n", ver);
			put_str ("N - See what's new\n");
			put_str ("D - Download the new version to your home "
					"directory\n");
			put_str ("Q - Quit this screen\n");
		}
		else {
			put_str ("\nYou have the latest version of the MOC, "
					"thanks for checking!.\n");
			put_str ("Press space.");
		}

		free (ver);
	}
}
